import { forwardRef, createElement } from 'react';

const table = (bool, number, string, array, object, fn) => ({
    [Boolean.toString()]: bool,
    [Number.toString()]: number,
    [String.toString()]: string,
    [Array.toString()]: array,
    [Object.toString()]: object,
    [Function.toString()]: fn
});
const operations = table((value) => value, (value) => value, (value) => value, (value) => useClassName(...value), (value) => Object.keys(value).filter(key => useClassName(value[key])), (value) => useClassName(value()));
const type = (value) => (value.constructor.toString());
const pipe = (...fns) => (...args) => fns.reduce((acc, next) => next(acc), args);
const filter = (values) => values.filter(Boolean);
const map = (values) => values.map(value => (operations[type(value)] || (() => value))(value));
const flatten = (values) => (values.reduce((acc, val) => Array.isArray(val)
    ? acc.concat(flatten(val))
    : acc.concat(val), []));
const join = (values) => values.join(' ');
const trim = (values) => values.trim();
const useClassName = pipe(filter, map, flatten, join, trim);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

const capture = (target, accessed) => {
    const captured = {};
    let released = false;
    const spy = (target, prop) => {
        !released && accessed.add(prop);
        return target[prop];
    };
    const release = () => {
        released = true;
    };
    for (const prop in target) {
        if (!target.hasOwnProperty(prop)) {
            continue;
        }
        define(captured, prop, {
            enumerable: true,
            get: () => spy(target, prop),
        });
    }
    return {
        captured,
        release
    };
};
const define = (target, prop, attributes) => (Object.defineProperty(target, prop, attributes));
const sanitize = (keys, props) => {
    const remove = (_a, key) => {
        var _b = key, value = _a[_b], rest = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
        return rest;
    };
    return keys.reduce(remove, props);
};

const Type = {
    TYPE: 'type',
    PROPS: 'props'
};
const prepareHooks = (fns) => {
    const map = {};
    for (const fn of fns) {
        if (!fn) {
            continue;
        }
        const { type, hook } = fn;
        if (!map[type]) {
            map[type] = [];
        }
        map[type].push((args) => hook(args));
    }
    return map;
};
const interceptHooks = (fns) => {
    const map = {};
    for (const fn of fns) {
        if (!fn) {
            continue;
        }
        const { name, hook } = fn;
        map[name] = hook;
    }
    return (name) => {
        if (!map[name]) {
            return () => undefined;
        }
        return map[name];
    };
};
const useHooks = (use, context) => {
    if (!use) {
        return context;
    }
    const prepared = prepareHooks(use);
    for (const prop of Object.values(Type)) {
        const fns = prepared[prop];
        if (!fns) {
            continue;
        }
        for (const fn of fns) {
            const result = fn(context[prop]);
            if (typeof result !== "object") {
                context[prop] = result;
            }
            else {
                for (const key in result) {
                    if (!result.hasOwnProperty(key)) {
                        continue;
                    }
                    define(context[prop], key, {
                        enumerable: true,
                        value: result[key]
                    });
                }
            }
        }
    }
    return context;
};
const createHook = (type) => (name, hook) => ({ name, type, hook });
var hook = createHook(Type.PROPS);

var useType = (type) => createHook(Type.TYPE)('type', () => (type));

const factory = ({ type = null, className: baseClass = null, style = null, extend = (props) => (props), render = ({ children }) => (children) }) => {
    const build = (props, ref) => {
        let { className } = props;
        const keys = new Set(['use']);
        const { captured, release } = capture(props, keys);
        const classes = (style && style(captured)) || null;
        const extended = (extend && extend(captured)) || null;
        const children = (render && render(captured)) || null;
        release();
        if (!type) {
            return children;
        }
        className = useClassName(baseClass, classes, className);
        const { use = null } = Object.assign(Object.assign({}, props), extended);
        const sanitized = sanitize([...keys], Object.assign(Object.assign({}, props), extended));
        props = Object.assign(Object.assign({}, sanitized), { children,
            className,
            ref });
        const result = useHooks(use, { type, props });
        return createElement(result.type, result.props);
    };
    return forwardRef(build);
};

export { factory, hook, interceptHooks as intercept, useClassName, useType };
